// KeyFile.java
// Version 1.0 by DTAI - read old style keyboard file
// Version 2.0 by Larry S. Peranich, Nov. 25, 1998 - reads new style keyboard file
//             almost total re-write
//     Dec. 10, 1998 fixed Symbol file to use symCodes starting at 1, not zero,
//     so we can remove the kludge from Nov. 25 that subtracted 1 from symCode here
//     while reading

package signwriting;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;

/**
 * reads the keyboard file, and stores the data for recall by the scan code of
 * user keystrokes.  See KeyRep, SignRep, and MultiSignRep classes.
 */
public class KeyFile {

    public final static int Z     = 0x2C;
    public final static int X     = 0x2D;
    public final static int C     = 0x2E;
    public final static int V     = 0x2F;
    public final static int B     = 0x30;
    public final static int N     = 0x31;
    public final static int M     = 0x32;
    public final static int Comma = 0x33;
    public final static int Dot   = 0x34;
    public final static int Slash = 0x35;

    public final static int A     = 0x1E;
    public final static int S     = 0x1F;
    public final static int D     = 0x20;
    public final static int F     = 0x21;
    public final static int G     = 0x22;
    public final static int H     = 0x23;
    public final static int J     = 0x24;
    public final static int K     = 0x25;
    public final static int L     = 0x26;
    public final static int Semi  = 0x27;
    public final static int Rquot = 0x28;

    public final static int Q     = 0x10;
    public final static int W     = 0x11;
    public final static int E     = 0x12;
    public final static int R     = 0x13;
    public final static int T     = 0x14;
    public final static int Y     = 0x15;
    public final static int U     = 0x16;
    public final static int I     = 0x17;
    public final static int O     = 0x18;
    public final static int P     = 0x19;
    public final static int Lbrac = 0x1A;
    public final static int Rbrac = 0x1B;
    public final static int Baksl = 0x2B;

    public final static int Lquot = 0x29;
    public final static int n1    = 0x02;
    public final static int n2    = 0x03;
    public final static int n3    = 0x04;
    public final static int n4    = 0x05;
    public final static int n5    = 0x06;
    public final static int n6    = 0x07;
    public final static int n7    = 0x08;
    public final static int n8    = 0x09;
    public final static int n9    = 0x0A;
    public final static int n0    = 0x0B;
    public final static int Minus = 0x0C;
    public final static int Equal = 0x0D;

    public final static int zbs  = 0x0E;// debug -- command key scan codes
    public final static int zret = 0x1C;
    public final static int zesc = 0x01;
    public final static int ztab = 0x0F;

    private final static int SCAN_CODE_COUNT = 256;
    
    private final static int MAX_KEY_FILE_INDEX = 49;
    
    // The indices in the KeyFile are Macintosh key codes (??), map to our
    // scan codes (above)
    private final static int[] MapKeyFileIndex2ScanCode = {
        A, S, D, F, H, G, Z, X,  // 0X00 - 0X07
        C, V, -1, B, Q, W, E, R,  // 0X08 - 0X0F  note 0X0A not used
        Y, T, n1, n2, n3, n4, n6, n5,  // OX10 - 0X17
        Equal, n9, n7, Minus, n8, n0, Rbrac, O,  //  0X18 - 0X1F
        U, Lbrac, I, P, -1, L, J, Rquot, //  0X20 - 0X27   note 0X24 not used
        K, Semi, -1, Comma, Slash, N, M, Dot,  // 0X28 - 0X2F  note 0X2A not used
        -1, -1  // for unused two last key slots
    };

    String filename;
    static SignRep[] fingerKeyboard;  // stores symbols for each finger spelling sign
    static MultiSignRep[] signKeyboard; // stores sign variations each sign key represents
    
    /**
     * constructor, takes a symbol file name, opens it and reads it
     */
    public KeyFile() {
        filename = Resources.getResources().getString("KeyFile.file");
    }

    public void readKeys() throws IOException {
        int num_this_key;  // how many bytes
        int num_bytes_total = 0;
        int descender;
        int numSyms;
        int[] symcode = new int[5];
        int symbol_code;
        int var, rot, fill, x, y;
        boolean flop, mark1, mark2, mark3;
        int scan_code;

        DataInputStream in = new DataInputStream(
                             new BufferedInputStream(
                             new FileInputStream(filename)));

        /** Initialize keyboard arrays **/
        fingerKeyboard = new SignRep[SCAN_CODE_COUNT];
        signKeyboard = new MultiSignRep[SCAN_CODE_COUNT];
        for (int j = 0; j < SCAN_CODE_COUNT; j++) {
            fingerKeyboard[j] = null;
            signKeyboard[j] = null;
        }

        // Read in Finger Spelling keys
        for (int keycode = 0; keycode < MAX_KEY_FILE_INDEX + 1; keycode ++) {
            // get scan code for this keycode
            scan_code = MapKeyFileIndex2ScanCode[keycode];
            
            num_this_key = in.readUnsignedShort();  // number of bytes
            
            num_bytes_total += num_this_key + 2;
            if (num_this_key > 0) {
                descender = in.readUnsignedByte();
                numSyms = in.readUnsignedByte();
                
                // create SignRep object and put in keyboard array
                fingerKeyboard[scan_code] = new SignRep(numSyms);
                
                // read in each symbol, and put it in SignRep
                for (int i = 0; i < numSyms; i++)  {
                    for (int j = 0; j < 5; j++) {
                        symcode[j] = in.readUnsignedByte();
                    } // for j
                    
                    try {
                        mark1 = (symcode[0] & 0x80) != 0 ? true : false;
                        symbol_code = (symcode[0] & 0x7f) << 3 | (symcode[1] & 0xe0) >> 5;
                        var = (symcode[1] & 0x1c) >> 2;
                        flop = (symcode[1] & 0x02) != 0 ? true : false;
                        mark2 = (symcode[1] & 0x01) != 0 ? true : false;
                        mark3 = (symcode[4] & 0x80) != 0 ? true : false;
                        rot = (symcode[4] & 0x38) >> 3;
                        fill = (symcode[4] & 0x07);
                        x = (symcode[2] & 0xff) << 2 | (symcode[3] & 0xc0) >> 6;
                        y = (symcode[3] & 0x3f) << 1 | (symcode[4] & 0x40) >> 6;
                        if (mark1 || mark2 || mark3)  {
                            System.err.println("FingerSpelling Keyboard index " + keycode);
                            throw new KeyFileException("ERROR: At least one mark set: " + 
                                                    mark1 + mark2 + mark3);
                        }
                        if (var < 0 || var > 5)  {
                            System.err.println("FingerSpelling Keyboard index " + keycode);
                            throw new KeyFileException("ERROR: var out of range");
                        }
                        if (rot < 0 || rot > 7) {
                            System.err.println("FingerSpelling Keyboard index " + keycode);
                            throw new KeyFileException("ERROR: rot out of range");
                        }
                        if (fill < 0 || fill > 5) {
                            System.err.println("FingerSpelling Keyboard index " + keycode);
                            throw new KeyFileException("ERROR: fill out of range");
                        }
                    
                        // got a (hopefully) good symbol, put it in SignRep
                        fingerKeyboard[scan_code].addSymbol(symbol_code, rot, var,
                            fill, flop, x, y);
                    }
                    catch (KeyFileException kfe) {
                        System.err.println("FingerSpelling Keyboard index " + keycode +
                            " Symbol " + i +
                            "\nGot bad symbol data\n");
                        kfe.printStackTrace();
                        in.close();
                        System.exit(1);
                    }
                    catch (KeyNumSymException knse) {
                        System.err.println("FingerSpelling Keyboard index " + keycode +
                            " Symbol " + i +
                            "\nTrying to add too many symbols, ignoring excess\n");
                    }
                    catch (KeyErrorException kee) {
                        System.err.println("FingerSpelling Keyboard index " + keycode +
                            " Symbol " + i +
                            "\nCould not allocate SignSymbol\nWill fill in null pointer");
                        System.err.println(kee.getMessage() + "\n\n");
                        //kee.printStackTrace();
                        //in.close();
                        //System.exit(1);
                    }
                }  // for i
                
                // verify we have a good SignRep
                if (!fingerKeyboard[scan_code].isValid()) {
                    System.err.println("FingerSpelling Keyboard index " + keycode +
                            "Invalid SignRep after reading all for this index\n");
                    in.close();
                    System.exit(1);
                }
            }  // if num_this_key
        }  // for keycode
        

        // read in sign keyboard
        for (int keycode = 0; keycode < MAX_KEY_FILE_INDEX + 1; keycode ++) {
            // get scan code for this keycode
            scan_code = MapKeyFileIndex2ScanCode[keycode];
            
            num_this_key = in.readUnsignedShort();  // number of bytes
            num_bytes_total += num_this_key + 2;
            
            if (num_this_key > 0) {
                // create MultiSignRep for this key
                signKeyboard[scan_code] = new MultiSignRep(10);  // always 10 possible
                
                // read in data for each variation
                for (int keyvar = 0; keyvar < 10; keyvar++) {
                    descender = in.readUnsignedByte();
                    numSyms = in.readUnsignedByte();
                    
                    if (numSyms == 0) {
                        // stick an empty SignRep in this slot
                        try {
                            signKeyboard[scan_code].addSymbol(keyvar, 0, 
                                0, 0, 0, 0, false, 0, 0);
                        }
                        catch (Exception e) {
                            System.err.println("Sign Keyboard index " + keycode +
                                " variation " + keyvar + 
                                "\nCould not put empty sign in keyboard\n");
                                e.printStackTrace();
                                in.close();
                                System.exit(1);
                        }
                        continue;
                    }
                    
                    // read in symbols for this variation
                    for (int i = 0; i < numSyms; i++)  {
                        for (int j = 0; j < 5; j++) {
                            symcode[j] = in.readUnsignedByte();
                        } // for j
                        try {
                            mark1 = (symcode[0] & 0x80) != 0 ? true : false;
                            symbol_code = (symcode[0] & 0x7f) << 3 | (symcode[1] & 0xe0) >> 5;
                            var = (symcode[1] & 0x1c) >> 2;
                            flop = (symcode[1] & 0x02) != 0 ? true : false;
                            mark2 = (symcode[1] & 0x01) != 0 ? true : false;
                            mark3 = (symcode[4] & 0x80) != 0 ? true : false;
                            rot = (symcode[4] & 0x38) >> 3;
                            fill = (symcode[4] & 0x07);
                            x = (symcode[2] & 0xff) << 2 | (symcode[3] & 0xc0) >> 6;
                            y = (symcode[3] & 0x3f) << 1 | (symcode[4] & 0x40) >> 6;
                            
                            // following is necessary to properly place symbols in
                            // multi-symbol sign
                            if (numSyms == 1) {
                                x = 0;
                                y = 0;
                            }
                            if (mark1 || mark2 || mark3)  {
                                System.err.println("Sign Keyboard index " + keycode +
                                    " variation " + keyvar);
                                throw new KeyFileException("ERROR: At least one mark set: " + 
                                    mark1 + mark2 + mark3);
                            }
                            if (var < 0 || var > 5) {
                                System.err.println("Sign Keyboard index " + keycode +
                                    " variation " + keyvar);
                                throw new KeyFileException("ERROR: var out of range");
                            }
                            if (rot < 0 || rot > 7)  {
                                System.err.println("Sign Keyboard index " + keycode +
                                    " variation " + keyvar);
                                throw new KeyFileException("ERROR: rot out of range");
                            }
                            if (fill < 0 || fill > 5) {
                                System.err.println("Sign Keyboard index " + keycode +
                                    " variation " + keyvar);
                                throw new KeyFileException("ERROR: fill out of range");
                            }
                            
                            // got a (hopefully) good symbol, put it in MultiSignRep
                            signKeyboard[scan_code].addSymbol(keyvar, numSyms,
                                    symbol_code, rot, var, fill, flop, x, y);
                        }
                        catch (KeyFileException kfe) {
                            System.err.println("Sign Keyboard index " + keycode +
                                " Symbol " + i +
                                "\nGot bad symbol data\n");
                            kfe.printStackTrace();
                            in.close();
                            System.exit(1);
                        }
                        catch (KeyNumSymException knse) {
                            System.err.println("Sign Keyboard index " + keycode +
                                " Symbol " + i +
                                "\nTrying to add too many symbols, ignoring excess\n");
                        }
                        catch (KeyErrorException kee) {
                            System.err.println("Sign Keyboard index " + keycode +
                                " variation " + keyvar + " Symbol " + i +
                                "\nCould not allocate SignSymbol\nPut in null pointer");
                            System.err.println(kee.getMessage() + "\n\n");
                            //in.close();
                            //System.exit(1);
                        }
                    }  // for i
                } // for keyvar
                // verify that we have a valid MultiSignRep
                if (!signKeyboard[scan_code].isValid()) {
                    System.err.println("Sign Keyboard index " + keycode +
                            "Invalid MultiSignRep after reading all for this index\n");
                    in.close();
                    System.exit(1);
                }
            }  // if num_this_key
        }  // for keycode
        
        // fill in all null spots in keyboards with empty SignRep's (MultiSignRep's)
        // they should never be hit, but . . . .
        for (int i = 0; i < SCAN_CODE_COUNT; i++) {
            if(fingerKeyboard[i] == null) fingerKeyboard[i] = new SignRep(0);
            if(signKeyboard[i] == null) signKeyboard[i] = new MultiSignRep(0);
        }
        in.close();
    }  // readKeys()
    
    public static int getScanCode(char keycode) {
        switch (keycode) {
            case '`': return Lquot;
            case '1': return n1;
            case '2': return n2;
            case '3': return n3;
            case '4': return n4;
            case '5': return n5;
            case '6': return n6;
            case '7': return n7;
            case '8': return n8;
            case '9': return n9;
            case '0': return n0;
            case '-': return Minus;
            case '=': return Equal;

            case '~': return Lquot;
            case '!': return n1;
            case '@': return n2;
            case '#': return n3;
            case '$': return n4;
            case '%': return n5;
            case '^': return n6;
            case '&': return n7;
            case '*': return n8;
            case '(': return n9;
            case ')': return n9;
            case '_': return Minus;
            case '+': return Equal;

            case 'q': return Q;
            case 'w': return W;
            case 'e': return E;
            case 'r': return R;
            case 't': return T;
            case 'y': return Y;
            case 'u': return U;
            case 'i': return I;
            case 'o': return O;
            case 'p': return P;
            case '[': return Lbrac;
            case ']': return Rbrac;
            case '\\': return Baksl;

            case 'Q': return Q;
            case 'W': return W;
            case 'E': return E;
            case 'R': return R;
            case 'T': return T;
            case 'Y': return Y;
            case 'U': return U;
            case 'I': return I;
            case 'O': return O;
            case 'P': return P;
            case '{': return Lbrac;
            case '}': return Rbrac;
            case '|': return Baksl;

            case 'a': return A;
            case 's': return S;
            case 'd': return D;
            case 'f': return F;
            case 'g': return G;
            case 'h': return H;
            case 'j': return J;
            case 'k': return K;
            case 'l': return L;
            case ';': return Semi;
            case '\'': return Rquot;

            case 'A': return A;
            case 'S': return S;
            case 'D': return D;
            case 'F': return F;
            case 'G': return G;
            case 'H': return H;
            case 'J': return J;
            case 'K': return K;
            case 'L': return L;
            case ':': return Semi;
            case '\"': return Rquot;

            case 'z': return Z;
            case 'x': return X;
            case 'c': return C;
            case 'v': return V;
            case 'b': return B;
            case 'n': return N;
            case 'm': return M;
            case ',': return Comma;
            case '.': return Dot;
            case '/': return Slash;

            case 'Z': return Z;
            case 'X': return X;
            case 'C': return C;
            case 'V': return V;
            case 'B': return B;
            case 'N': return N;
            case 'M': return M;
            case '<': return Comma;
            case '>': return Dot;
            case '?': return Slash;

            case 8 : return zbs;
            case 13: return zret;
            case 27: return zesc;
            case 9 : return ztab;

            default: return 255;
        }
    }
    
    public static int getKeynum(int scanCode) {
	    int keynum = -1;
	    switch(scanCode) {
	        case A:
	            keynum = 0;
	            break;
	        case S:
	            keynum = 1;
	            break;
	        case D:
	            keynum = 2;
	            break;
	        case F:
	            keynum = 3;
	            break;
	        case J:
	            keynum = 4;
	            break;
	        case K:
	            keynum = 5;
	            break;
	        case L:
	            keynum = 6;
	            break;
	        case Semi:
	            keynum = 7;
	            break;
	        case G:
	            keynum = 8;
	            break;
	        case H:
	            keynum = 9;
	            break;
	    }
	    return keynum;
	}
	
}

class KeyFileException extends Exception {
    public KeyFileException(String s) {
        super(s);
    }
}

